"""SYML (Simple YAML-like Markup Language) is a simple markup language with similar structure to YAML, but without all the gewgaws and folderol."""

from io import TextIOBase
from typing import Any

from . import parsers
from .basetypes import StrPath


def loads(document: str, filename: StrPath | None = None) -> list[Any] | dict[str, Any] | str:
    """Load a SYML document from a string."""
    return parsers.parse(document, filename=filename).as_data()


def load(file_obj: TextIOBase, filename: StrPath | None = None) -> list[Any] | dict[str, Any] | str:
    """Load a SYML document from a file-like object."""
    return loads(file_obj.read(), filename=filename)



"""Utility functions for SYML"""

from functools import cache


@cache
def split_lines(text: str, keepends: bool = False) -> list[str]:  # noqa: FBT001, FBT002
    """Memoized version of str.splitlines()"""
    return text.splitlines(keepends=keepends)


def get_line(text: str, line_number: int) -> str:
    """Return the contents of the specified line number from the given text."""
    try:
        return split_lines(text, keepends=True)[line_number - 1]
    except IndexError:
        return ''



"""SYML parsers"""

from __future__ import annotations

import textwrap
from typing import TYPE_CHECKING

from parsimonious import Grammar, NodeVisitor

from . import nodes
from .exceptions import OutOfContextNodeError

if TYPE_CHECKING:  # pragma: nocover
    from parsimonious.nodes import Node as PNode

    from .basetypes import StrPath
    from .nodes import OptionalNodes, OptionalSymlNodes, SymlNode, SymlNodes


class SymlParser(NodeVisitor):  # type: ignore[type-arg]
    """Parser for SYML"""

    grammar = Grammar(
        textwrap.dedent(
            r"""
            lines       = line*
            line        = indent (comment / blank / structure / value) &eol
            structure   = list_item / key_value / section

            indent      = ~"\s*"

            blank       = &eol
            comment     = ~"(#|//+)+" text?

            list_item   = "-" ws value

            key_value   = section ws data
            section     = key ":"
            key         = ~"[^\s:]+"

            eol         = "\n" / ~"$"
            ws          = ~"[ \t]+"
            text        = ~".+"

            value       = structure / data
            data        = text

            """
        )
    )
    unwrapped_exceptions = (OutOfContextNodeError,)

    def __init__(self, filename: StrPath | None = None) -> None:
        super().__init__()
        self.filename = filename

    def reduce_children(self, children: OptionalSymlNodes) -> SymlNodes:
        """Return all non-null children."""
        return [c for c in children if c is not None]

    def visit_blank(self, node: PNode, children: SymlNodes) -> None:  # noqa: ARG002
        """Visit a blank."""
        return

    def visit_line(self, node: PNode, children: SymlNodes) -> OptionalNodes:  # noqa: ARG002
        """Visit a line."""
        indent, value, _ = children
        if value is not None:
            value.set_level(indent.level)  # type: ignore[arg-type]
            return value
        return None

    def generic_visit(self, node: PNode, children: OptionalSymlNodes) -> SymlNodes | SymlNode | None:  # type: ignore[override]  # noqa: ARG002
        """Visit a generic node."""
        nodes = self.reduce_children(children)
        if not nodes:
            return None
        if len(nodes) == 1:
            return nodes[0]
        else:  # pragma: nocover  # noqa: RET505
            return nodes

    def visit_text(self, node: PNode, children: SymlNodes) -> nodes.TextLeafNode:  # noqa: ARG002
        """Return a text leaf node."""
        return nodes.TextLeafNode(pnode=node, filename=self.filename)

    def visit_key(self, node: PNode, children: SymlNodes) -> nodes.KeyLeafNode:  # noqa: ARG002
        """Return a key leaf node."""
        return nodes.KeyLeafNode(pnode=node, filename=self.filename)

    def visit_comment(self, node: PNode, children: SymlNodes) -> nodes.Comment:  # noqa: ARG002
        """Visit a comment node."""
        _, text = children
        return nodes.Comment(pnode=text.pnode, filename=self.filename)

    def visit_indent(self, node: PNode, children: SymlNodes) -> nodes.IndentNode:  # noqa: ARG002
        """Visit an indentation token."""
        return nodes.IndentNode(
            pnode=node, level=len(node.text.replace('\t', ' ' * 4).strip('\n')), filename=self.filename
        )

    def visit_key_value(self, node: PNode, children: SymlNodes) -> OptionalNodes:  # noqa: ARG002
        """Visit a mapping value."""
        section, _, value = children
        section.incorporate_node(value)
        return section

    def visit_section(self, node: PNode, children: SymlNodes) -> nodes.KeyValue:
        """Visit a key/value section."""
        key, _ = children
        return nodes.KeyValue(pnode=node, key=key, filename=self.filename)  # type: ignore[arg-type]

    def visit_list_item(self, node: PNode, children: SymlNodes) -> nodes.ListItem:
        """Visit a list item."""
        _, _, value = children
        li = nodes.ListItem(pnode=node, filename=self.filename)
        if value is not None:  # pragma: nobranch
            li.incorporate_node(value)
        return li

    def visit_lines(self, node: PNode, children: OptionalSymlNodes) -> nodes.Root:
        """Visit the lines within a SYML document."""
        root = nodes.Root(pnode=node, filename=self.filename)
        current: SymlNode = root

        for child in self.reduce_children(children):
            if isinstance(child, nodes.Comment):
                current.comments.append(child)
            else:
                current = current.incorporate_node(child)
        return root


def parse(source_syml: str, filename: StrPath | None = None) -> nodes.Root:
    """Parse a SYML document."""
    return SymlParser(filename=filename).parse(source_syml)



"""Node implementations for parsing SYML documents"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:  # pragma: nocover
    from parsimonious.nodes import Node as PNode

from .basetypes import Pos, Source, StrPath
from .exceptions import OutOfContextNodeError
from .utils import get_line


@dataclass(kw_only=True)
class SymlNode:
    """A generic node in a SYML document."""

    pnode: PNode = field(repr=False)
    level: int | None = field(default=None)
    parent: SymlNode | None = field(default=None)
    comments: list[Comment] = field(default_factory=list)
    children: list[SymlNode] = field(default_factory=list)
    filename: StrPath | None = field(default=None)

    source: Source = field(init=False)

    def __post_init__(self) -> None:
        self.source = Source.from_node(self.pnode, filename=self.filename)
        if self.level is not None:
            self.set_level(self.level)

    def set_level(self, level: int) -> None:
        """Set this node's level."""
        self.level = level
        for child in self.children:
            child.set_level(level)

    def as_data(self) -> Any:  # noqa: ANN401  # pragma: nocover
        """Return this node as primitive data types."""
        raise NotImplementedError

    def as_source(self) -> Any:  # noqa: ANN401  # pragma: nocover
        """Return this node as primitive data types with Source objects for strings."""
        raise NotImplementedError

    def get_tip(self) -> SymlNode:
        """Return the tip of this branch."""
        if self.children:
            return self.children[-1].get_tip()
        return self

    def can_add_node(self, node: SymlNode) -> bool:  # noqa: ARG002  # pragma: nocover
        """Check if this node can add a child node."""
        return False

    def add_node(self, node: SymlNode) -> SymlNode:
        """Add a child node."""
        self.children.append(node)
        node.parent = self
        if node.level is None:
            node.set_level(self.level)  # type: ignore[arg-type]
        return node.get_tip()

    def incorporate_node(self, node: SymlNode) -> SymlNode:
        """Incorporate the given node into the tree somewhere nearby."""
        if self.can_add_node(node):
            return self.add_node(node)
        if self.parent is not None:
            return self.parent.incorporate_node(node)
        else:  # pragma: nocover  # noqa: RET505
            # Shouldn't ever get here:
            self.fail_to_incorporate_node(node)
            return self  # pragma: nocover

    def fail_to_incorporate_node(self, node: SymlNode) -> None:
        """Report a failure to incorporate a node."""
        pnode = node.pnode
        pos = Pos.from_str_index(pnode.full_text, pnode.start)
        line = get_line(pnode.full_text, pos.line)
        raise OutOfContextNodeError('Failed to incorporate a node', pos, line)


class IndentNode(SymlNode):
    """A node representing an indentation."""


SymlNodes = list[SymlNode]
OptionalSymlNodes = list[SymlNode | None]

NodeOrNodes = SymlNodes | SymlNode | str

OptionalNodes = NodeOrNodes | None


@dataclass(kw_only=True)
class ContainerNode(SymlNode):
    """A container node that may contain a child value."""

    def as_source(self) -> Any:  # noqa: ANN401  # pragma: nocover
        """Return this node as primitive data types with Source objects for strings."""
        if self.children:
            return self.children[0].as_source()
        return None  # pragma: nocover

    def as_data(self) -> Any:  # noqa: ANN401
        """Return the container as primitive data types."""
        if self.children:
            return self.children[0].as_data()
        return None  # pragma: nocover

    def incorporate_node(self, node: SymlNode) -> SymlNode:
        """Incorporate the given node into this branch."""
        if self.can_add_node(node):
            intermediary: SymlNode | None = None
            if isinstance(node, KeyValue):
                intermediary = Mapping(pnode=node.pnode, level=self.level, filename=self.filename)
            elif isinstance(node, ListItem):
                intermediary = List(pnode=node.pnode, level=self.level, filename=self.filename)

            if intermediary is not None:
                intermediary.set_level(node.level)  # type: ignore[arg-type]
                intermediary = self.incorporate_node(intermediary)
                return intermediary.incorporate_node(node)
            return super().incorporate_node(node)
        if self.parent is not None:
            return self.parent.incorporate_node(node)
        else:  # pragma: nocover  # noqa: RET505
            self.fail_to_incorporate_node(node)
            return self

    def can_add_node(self, node: SymlNode) -> bool:
        """Check if this container can add a child node."""
        return not self.children and (node.level is None or (self.level is not None and node.level > self.level))


@dataclass(kw_only=True)
class ParentNode(SymlNode):
    """A parent node that con have multiple children"""

    def can_add_node(self, node: SymlNode) -> bool:
        """Check if a child node can be added."""
        return node.level is None or (self.level is not None and node.level >= self.level)

    def add_node(self, node: SymlNode) -> SymlNode:
        """Add a child node."""
        self.children.append(node)
        node.parent = self
        if node.level is None:
            node.set_level(self.level)  # type: ignore[arg-type]
        return node.get_tip()


@dataclass(kw_only=True)
class Root(ContainerNode):
    """A root container node for a SYML document"""

    level: int = field(default=0)

    def can_add_node(self, node: SymlNode) -> bool:
        """Check if a child node may be added."""
        return not self.children and (node.level is None or (self.level is not None and node.level >= self.level))


class List(ParentNode):
    """A list node"""

    def can_add_node(self, node: SymlNode) -> bool:
        """Check if a child node can be added."""
        return super().can_add_node(node) and isinstance(node, ListItem)

    def as_source(self) -> Any:  # noqa: ANN401  # pragma: nocover
        """Return this node as primitive data types with Source objects for strings."""
        return [c.as_source() for c in self.children]

    def as_data(self) -> list[Any]:
        """Return this node as primitive data types."""
        return [c.as_data() for c in self.children]


class ListItem(ContainerNode):
    """A list item within a list."""


class Mapping(ParentNode):
    """A mapping of keys to values"""

    def can_add_node(self, node: SymlNode) -> bool:
        """Check if a child node may be added."""
        return super().can_add_node(node) and isinstance(node, KeyValue)

    def as_source(self) -> Any:  # noqa: ANN401  # pragma: nocover
        """Return this node as primitive data types with Source objects for strings."""
        return {c.key.as_source(): c.as_source() for c in self.children}  # type: ignore[attr-defined]

    def as_data(self) -> dict[str, Any]:
        """Return this node as primitive data types."""
        return {c.key.as_data(): c.as_data() for c in self.children}  # type: ignore[attr-defined]


@dataclass(kw_only=True)
class KeyValue(ContainerNode):
    """A key-value item within a mapping"""

    key: KeyLeafNode


@dataclass(kw_only=True)
class TextLeafNode(SymlNode):
    """A leaf node containing a text value."""

    def as_source(self) -> Source:  # pragma: nocover
        """Return this node as primitive data types with Source objects for strings."""
        source = self.source
        for child in self.children:
            source += child.as_source()
        return source

    def as_data(self) -> str:
        """Return this node as primitive types."""
        return '\n'.join([str(self.source)] + [c.as_data() for c in self.children])

    def can_add_node(self, node: SymlNode) -> bool:
        """Check if a child node can be added."""
        return isinstance(node, TextLeafNode)


class KeyLeafNode(SymlNode):
    """A leaf node containing a key value."""

    def can_add_node(self, node: SymlNode) -> bool:  # noqa: ARG002  # pragma: nocover
        """Check if this node can add a child. It can't."""
        return False

    @property
    def key(self) -> Source:
        """Return a Source object representing the key."""
        return Source.from_node(self.pnode, filename=self.filename)

    def as_source(self) -> Any:  # noqa: ANN401  # pragma: nocover
        """Return this node as primitive data types with Source objects for strings."""
        return self.key

    def as_data(self) -> str:
        """Return the key as a string."""
        return str(self.key)


class Comment(TextLeafNode):
    """A comment node"""

    def as_data(self) -> str:  # pragma: nocover
        """Return an empty string."""
        return ''

    def can_add_node(self, node: SymlNode) -> bool:  # pragma: nocover
        """Check if a child node can be added."""
        return isinstance(node, Comment)



"""SYML base types"""

from __future__ import annotations

import re
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from syml import utils

if TYPE_CHECKING:  # pragma: nocover
    from parsimonious.nodes import Node as PNode


StrPath = str | Path


@dataclass(slots=True, frozen=True)
class Pos:
    """A position within a source file"""

    index: int
    line: int
    column: int

    @classmethod
    def from_str_index(cls, text: str, index: int) -> Pos:
        """Get (line_number, col) of `index` in `string`.

        Based on http://stackoverflow.com/a/24495900
        """
        lines = utils.split_lines(text, keepends=True)
        curr_pos = 0
        linenum = 0
        for linenum, line in enumerate(lines):
            if curr_pos + len(line) > index:
                return cls(index, linenum + 1, index - curr_pos)
            curr_pos += len(line)
        return cls(len(text), linenum + 1, 0)


@dataclass(slots=True, repr=False, frozen=True)
class Source:
    """A line within a source file"""

    filename: StrPath | None
    start: Pos
    end: Pos
    text: str

    @classmethod
    def from_node(cls, pnode: PNode, filename: StrPath | None = None) -> Source:
        """Build a Source from the given PNode, filename, and line value."""
        return cls(
            filename=filename,
            start=Pos.from_str_index(pnode.full_text, pnode.start),
            end=Pos.from_str_index(pnode.full_text, pnode.end),
            text=pnode.text,
        )

    @classmethod
    def from_text(
        cls,
        text: str,
        substring: str | None = None,
        source_text: str | None = None,
        filename: StrPath | None = None,
    ) -> Source:
        """Build a Source from the given components."""
        if substring is None:  # pragma: no cover
            substring = text
        match = re.search(substring, text)
        if match is None:
            raise ValueError('No match found', substring)
        source_text = match.group() if source_text is None else source_text
        return Source(
            filename=filename,
            start=Pos.from_str_index(text, match.start()),
            end=Pos.from_str_index(text, match.end()),
            text=source_text,
        )

    def __repr__(self) -> str:
        filename = f'{self.filename}, ' if self.filename else ''
        return f'<Source: {filename}Line {self.start.line}, Column {self.start.column} (index {self.start.index}): {self.text!r}>'

    def __str__(self) -> str:
        return self.text

    def __add__(self, other: SourceStr) -> Source:
        if isinstance(other, str):
            lines = utils.split_lines(other, keepends=True)
            return Source(
                filename=self.filename,
                start=self.start,
                end=Pos(index=self.end.index + len(other), line=self.end.line + len(lines), column=len(lines[-1])),
                text=f'{self.text}\n{other}',
            )
        if isinstance(other, Source):
            return Source(
                filename=self.filename,
                start=self.start,
                end=other.end,
                text=f'{self.text}\n{other}',
            )
            return self + other.text

        raise TypeError('Tried to add invalid type to Source', type(other))

    def __hash__(self) -> int:  # pragma: no cover
        return hash(str(self))


SourceStr = Source | str



"""Exceptions for parsing SYML documents"""


class ParseError(ValueError):
    """An error encountered while parsing"""


class OutOfContextNodeError(ParseError):
    """A node encountered in an illegal context"""
